/*
 NoiascaBufferPrint.cpp based on:
 
 Print.cpp - Base class that provides print() and println()
 Copyright (c) 2008 David A. Mellis.  All right reserved.
 
 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.
 
 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.
 
 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
 Modified 23 November 2006 by David A. Mellis
 Modified 03 August 2015 by Chuck Todd
 
 Modified 2021-10-28 by noiasca
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "Arduino.h"

#include "NoiascaBufferPrint.h"

// Public Methods //////////////////////////////////////////////////////////////

/* default implementation: may be overridden */
size_t NoiascaBufferPrint::write(const uint8_t *buffer, size_t size)
{
  size_t n = 0;
  while (size--) {
    if (write(*buffer++)) n++;
    else break;
  }
  return n;
}

/* noiasca - based on write */
size_t NoiascaBufferPrint::add(const uint8_t *buffer, size_t size)
{
  size_t n = 0;
  while (size--) {
    if (add(*buffer++)) n++;
    else break;
  }
  return n;
}

size_t NoiascaBufferPrint::getBufferLength()
{
  return actual;
}

/**
      \param newValue the new actual position in the buffer
*/
int NoiascaBufferPrint::setBufferLength(size_t newValue)
{
  actual = newValue;
  return 0;        // success
}

int NoiascaBufferPrint::resetBuffer()
{
  setBufferLength(0);
  return 0;        // success
}

/**
      \param newValue the new limit to excecute a auto flush
*/
int NoiascaBufferPrint::setFlushAfter(size_t newValue)
{
  if (newValue < buffersize)
  {
    flushAfter = newValue; 
    return 0;      // success
  } 
  else 
  {
    return 1;      // error
  }
}

// noiasca adopted functions
size_t NoiascaBufferPrint::print(const __FlashStringHelper *ifsh, bool finalFlush)
{
  PGM_P p = reinterpret_cast<PGM_P>(ifsh);
  size_t n = 0;
  while (1) {
    unsigned char c = pgm_read_byte(p++);
    if (c == 0) break;
    if (add(c)) n++;
    else break;
  }
  if(finalFlush) flush();
  return n;
}

size_t NoiascaBufferPrint::print(const String &s, bool finalFlush)
{
  size_t n = add(s.c_str(), s.length());
  if(finalFlush | (actual >= flushAfter)) flush();
  return n;
}

size_t NoiascaBufferPrint::print(const char str[], bool finalFlush)
{
  size_t n = add(str);
  if(finalFlush | (actual >= flushAfter)) flush();
  return n;
}

size_t NoiascaBufferPrint::print(char c, bool finalFlush)
{
  size_t n = add(c);
  if(finalFlush | (actual >= flushAfter)) flush();
  return n;
}

size_t NoiascaBufferPrint::print(unsigned char b, int base, bool finalFlush)
{
  size_t n = print((unsigned long) b, base, false);
  if(finalFlush | (actual >= flushAfter)) flush();
  return n;
}

/**
      \param n the integer
      \param base format DEC, BIN, OCT or HEX
      \param finalFlush autoflush after last digit
*/
size_t NoiascaBufferPrint::print(int n, int base, bool finalFlush)
{
  size_t result = print((long) n, base, false);
  if(finalFlush | (actual >= flushAfter)) flush();
  return result; 
}

size_t NoiascaBufferPrint::print(unsigned int n, int base, bool finalFlush)
{
  size_t result = print((unsigned long) n, base, false);
  if(finalFlush | (actual >= flushAfter)) flush();
  return result;
}

size_t NoiascaBufferPrint::print(long n, int base, bool finalFlush)               
{
  size_t result = 0;
  if (base == 0) {
    result = add(n);
  } else if (base == 10) {
    size_t t = 0;
    if (n < 0) {
      t = add('-');                 // not print but add
      n = -n;
    }
    result = printNumber(n, 10) + t;
  } else {
    result = printNumber(n, base);
  }
  if(finalFlush | (actual >= flushAfter)) flush();
  return result;
}

size_t NoiascaBufferPrint::print(unsigned long n, int base, bool finalFlush)
{
  size_t result;
  if (base == 0) result = add(n);
  else result =  printNumber(n, base);
  if(finalFlush | (actual >= flushAfter)) flush();
  return result; 
}

size_t NoiascaBufferPrint::print(double n, int digits, bool finalFlush)
{
  size_t result = printFloat(n, digits);
  if(finalFlush | (actual >= flushAfter)) flush();
  return result;
}

// no printable in this class
//size_t NoiascaBufferPrint::print(const Printable& x)
//{
//  return x.printTo(*this);
//}


/*
   all println(with parameter) end with a println().
   That gives two possibilites:
   
   a)
   Call the println() without parameter and use its
   implizit finalFlush (true) 
   instead of println(false) and a separate.
   
   b)
   call println(false) and use an explizit flush() to flush the buffer content.
   This enables one to add several println into the buffer. 
   You only have to set the finalflush to false and can continue to 
   fill the buffer.
*/

size_t NoiascaBufferPrint::println(const __FlashStringHelper *ifsh, bool finalFlush)
{
  size_t n = print(ifsh, false);
  n += println(false);
  if(finalFlush | (actual >= flushAfter)) flush();
  return n;
}

size_t NoiascaBufferPrint::println(bool finalFlush)
{
  size_t result = add("\r\n");
  if(finalFlush | (actual >= flushAfter)) flush();
  return result;
}

size_t NoiascaBufferPrint::println(const String &s, bool finalFlush)
{
  size_t n = print(s, false);
  n += println(false);
  if(finalFlush | (actual >= flushAfter)) flush();
  return n;
}

size_t NoiascaBufferPrint::println(const char c[], bool finalFlush)
{
  size_t n = print(c, false);
  n += println(false);
  if(finalFlush | (actual >= flushAfter)) flush();
  return n;
}

size_t NoiascaBufferPrint::println(char c, bool finalFlush)
{
  size_t n = print(c, false);
  n += println(false);
  if(finalFlush | (actual >= flushAfter)) flush();
  return n;
}

size_t NoiascaBufferPrint::println(unsigned char b, int base, bool finalFlush)
{
  size_t n = print(b, base, false);
  n += println(false);
  if(finalFlush | (actual >= flushAfter)) flush();
  return n;
}

size_t NoiascaBufferPrint::println(int num, int base, bool finalFlush)
{
  size_t n = print(num, base, false);
  n += println(false);
  if(finalFlush | (actual >= flushAfter)) flush();
  return n;
}

size_t NoiascaBufferPrint::println(unsigned int num, int base, bool finalFlush)
{
  size_t n = print(num, base, false);
  n += println(false);
  if(finalFlush | (actual >= flushAfter)) flush();
  return n;
}

size_t NoiascaBufferPrint::println(long num, int base, bool finalFlush)
{
  size_t n = print(num, base, false);
  n += println(false);
  if(finalFlush | (actual >= flushAfter)) flush();
  return n;
}

size_t NoiascaBufferPrint::println(unsigned long num, int base, bool finalFlush)
{
  size_t n = print(num, base, false);
  n += println(false);
  if(finalFlush | (actual >= flushAfter)) flush();
  return n;
}

size_t NoiascaBufferPrint::println(double num, int digits, bool finalFlush)
{
  size_t n = print(num, digits, false);
  n += println(false);
  if(finalFlush | (actual >= flushAfter)) flush();
  return n;
}

//size_t NoiascaBufferPrint::println(const Printable& x)
//{
//  size_t n = print(x);
//  n += println();
//  return n;
//}


// Private Methods /////////////////////////////////////////////////////////////

size_t NoiascaBufferPrint::printNumber(unsigned long n, uint8_t base)
{
  char buf[8 * sizeof(long) + 1]; // Assumes 8-bit chars plus zero byte.
  char *str = &buf[sizeof(buf) - 1];

  *str = '\0';

  // prevent crash if called with base == 1
  if (base < 2) base = 10;

  do {
    char c = n % base;
    n /= base;

    *--str = c < 10 ? c + '0' : c + 'A' - 10;
  } while(n);

  return add(str);
}

size_t NoiascaBufferPrint::printFloat(double number, uint8_t digits) 
{ 
  size_t n = 0;
  
  if (isnan(number)) return print("nan");
  if (isinf(number)) return print("inf");
  if (number > 4294967040.0) return print ("ovf");  // constant determined empirically
  if (number <-4294967040.0) return print ("ovf");  // constant determined empirically
  
  // Handle negative numbers
  if (number < 0.0)
  {
     n += print('-');
     number = -number;
  }

  // Round correctly so that print(1.999, 2) prints as "2.00"
  double rounding = 0.5;
  for (uint8_t i=0; i<digits; ++i)
    rounding /= 10.0;
  
  number += rounding;

  // Extract the integer part of the number and print it
  unsigned long int_part = (unsigned long)number;
  double remainder = number - (double)int_part;
  n += print(int_part);

  // Print the decimal point, but only if there are digits beyond
  if (digits > 0) {
    n += print('.'); 
  }

  // Extract digits from the remainder one at a time
  while (digits-- > 0)
  {
    remainder *= 10.0;
    unsigned int toPrint = (unsigned int)(remainder);
    n += print(toPrint);
    remainder -= toPrint; 
  } 
  
  return n;
}
