/*
   NoiascaBufferPrint.h based on:
  
  Print.h - Base class that provides print() and println()
  Copyright (c) 2008 David A. Mellis.  All right reserved.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
  Modified
  2022-04-17       some variables as protected
  2021-10-28 0.0.1 by noiasca
*/

/**
   \mainpage The Noiasca Buffer Print Library
   
   \section intro Introduction
   
   The Noiasca Buffer Print is a library very similar to Print.h but will buffer outputs until flush is called. 
   Despite other libraries, this class calls flush at the end of each print call. 
   The objective is to have an replacement for the Print.h which collects data and sends data to an interface not 
   bytewise with single writes but as one stream with several bytes.
   Like the Print.h library NoiascaBufferPrint.h is not a stand alone library, but can be used with other classes. 
      
   \section functionality How Noiasca Buffer Print work
   
   All print functions call the function .add instead of .write
   
   add() adds the value to the internal buffer.
   add() doesn't output - in other words add() doesn't call write(). 
   
   It is not needed that add() is implemented in the derived class
   
   flush() empties the buffer and streams to your class
   flush() needs to be implemented in the derived class

   write() exists similar to the Print.h library. 
   write() is used for the bytewise output to your class.
   write() doesn't use the internal buffer - but you could implement that in your class.
   write() needs to be implemented in the derived class (like with Print.h).
   
   print() and println() 
   have now an additional boolean parameter autoflush (default true).
   By default print() and println() will flush the buffer as last statement.
   If you want several print()/println() in the buffer, just add false.
   In this case - don't forget to flush().
   
   \subsection notes Some notes on the library behaviour
   There should be no need for an explicit flush() in your local write function because each print ends with a flush().
   But you could implement a trailing flush() in your write().
   
   If write() should be buffered also - implement it in the derived class.
   In this case don't forget to flush after write() also.
   
   Currently this library doesn't implement "printable" interface.
   
  \section implement_sec How to use Noiasca Buffer Print for your own classes
  When you define your class, don't inherit from Print.h but from NoiascaBufferPrint.h.
  You will need to implement two member functions in your class:
  + write() will be used for a single byte transmission. It is just a pass through from NoiascaBufferPrint to your class.
  It will look like the write() you would have used with Print.h
  + flush() will be called to transmit a stream after the end of transmission. 
  You can access the internal buffer with the member variable internal[].  
  I suggest to reset the internal buffer with resetBuffer() at last line of code of your flush() function.
  
  The example 01_hello_world shows a simple sketch using a buffered print to the serial interface.
  
   
  \section install_sec Installation
  
  \subsection step0 Download the library
  The newest version of this library can be downloaded from https://werner.rothschopf.net/microcontroller/202111_arduino_bufferprint.htm .
  
  \subsection step1 Install the library
 
  In the Arduino IDE use the Menu <br>
   Sketch / Include Library / Add .ZIP Library <br>
  to install the library.
*/

#ifndef  NoiascaBufferPrint_h
#define  NoiascaBufferPrint_h

#include <inttypes.h>
#include <stdio.h>           // for size_t

#include "WString.h"
//#include "Printable.h"     // printable not implemented in this library

#ifndef DEC                  // assumption: print.h was already included with other libraries
#define DEC 10
#endif
#ifndef HEX
#define HEX 16
#endif
#ifndef OCT
#define OCT 8
#endif
#ifndef BIN    
#define BIN 2
#endif

constexpr size_t buffersize = 64;      // size of the internal buffer
/**
  class to print into a buffer
*/
class NoiascaBufferPrint
{
  private:
    int write_error;
    size_t printNumber(unsigned long, uint8_t);
    size_t printFloat(double, uint8_t);
 
  protected:
    size_t flushAfter = buffersize;              // flush after n bytes in buffer
    size_t actual = 0;                           // actual write position of internal buffer
    void setWriteError(int err = 1) { write_error = err; }
/**
   \brief the internal buffered stream
   
   This array contains the buffered bytes - the stream to be outputed on next flush().
   Use this internal buffer together with getBufferLength() to evaluate the correct length of valid bytes.
*/
    char internal[buffersize];                    // internal buffer with stored characters
    
  public:
    NoiascaBufferPrint() : write_error(0) {}
    
    int getWriteError() { return write_error; }
    void clearWriteError() { setWriteError(0); }
  
  /**
    \brief similar functionality like in the default Print.h library. 
    
   write() is used for the bytewise output to your class.
   write() doesn't use the internal buffer - but you could implement that in your class.
   
   \note You must implement this function in your derived class (compareable with using Print.h).
   */
    virtual size_t write(uint8_t) = 0;
    
    size_t write(const char *str) {
      if (str == NULL) return 0;
      return write((const uint8_t *)str, strlen(str));
    }
    
    virtual size_t write(const uint8_t *buffer, size_t size);
    
    size_t write(const char *buffer, size_t size) {
      return write((const uint8_t *)buffer, size);
    }
    
    /**
      \brief puf a single value (one character) in to the internal buffer
      
      Adds the value to the internal buffer. 
      By default add() doesn't generate output. 
      \param value the value to be added to the buffer
    */
    size_t add(uint8_t value)
    {
      internal[actual] = value;
      actual++;
      if (actual >= buffersize) flush();
      return 1;
    }
    
    /*
      \param *str a pointer to the char array
    */
    size_t add(const char *str) {
      if (str == NULL) return 0;
      return add((const uint8_t *)str, strlen(str));
    }
    
    virtual size_t add(const uint8_t *buffer, size_t size);          // MISSING tbc if necessary
    
    size_t add(const char *buffer, size_t size) {
      return add((const uint8_t *)buffer, size);
    }

    // default to zero, meaning "a single write may block"
    // should be overriden by subclasses with buffering
    virtual int availableForWrite() { return 0; }

    size_t print(const __FlashStringHelper *, bool = true);
    size_t print(const String &, bool = true);
    size_t print(const char[], bool = true);
    size_t print(char, bool = true);
    size_t print(unsigned char, int = DEC, bool = true);
/**
    \brief print an integer to the buffer
    
    All print() and println() variants from the Print.h are available.
    All functions can take an additional parameter finalFlush which can be set to false to avoid flush after the print.
    
    - Call the println() without parameter and use its
      implizit finalFlush (true) 
      instead of println(false) and a separate.
    
    - Call println(false) and use an explizit flush() to flush the buffer content.
    
    This enables one to add several println into the buffer. 
    You only have to set the finalflush to false and can continue to 
    fill the buffer. 
    If the size of the buffer or the set limit is reached, print will flush the stream to your interface.      
*/
    size_t print(int, int = DEC, bool = true);
    size_t print(unsigned int, int = DEC, bool = true);
    size_t print(long, int = DEC, bool = true);
    size_t print(unsigned long, int = DEC, bool = true);
    size_t print(double, int = 2, bool = true);
//  size_t print(const Printable&);

    size_t println(const __FlashStringHelper *, bool = true);
    size_t println(const String &s, bool = true);
    size_t println(const char[], bool = true);
    size_t println(char, bool = true);
    size_t println(unsigned char, int = DEC, bool = true);
    size_t println(int, int = DEC, bool = true);
    size_t println(unsigned int, int = DEC, bool = true);
    size_t println(long, int = DEC, bool = true);
    size_t println(unsigned long, int = DEC, bool = true);
    size_t println(double, int = 2, bool = true);
//  size_t println(const Printable&);
    size_t println(bool = true);

/**
    \brief final method to flush data to the interface
    
    \note You must implement this function in your class.
*/
    virtual void flush() = 0;
    
/**
    \brief how many bytes are currently in the buffer
    
    Returns the actual bytes in the internal buffer. 
    Use this function in your implementation of flush() to determine, how many bytes should be streamed to your interface.
    \return actual bytes in buffer
*/
    size_t getBufferLength();

/**
    \brief set actual index
    
    sets the actual position in the buffer
    \return 0 on success
*/
    int setBufferLength(size_t);       
    
/**
    \brief reset the internal buffer to zero
    
    Call this function at the end of your implementation of add() to reset the internal buffer to 0.
    \return 0 on success
*/

    int resetBuffer();
/**
    \brief set value when to flush
    
    By default the internal buffer is 64 bytes and therefore the buffer will be flushed after 64 bytes latest.
    If you have an interface which can't handle a stream of 64 bytes, you can reduce the limit to your need.
    \return 0 on success
*/
    int setFlushAfter(size_t);         //  set actual/buffer position to a specific value
};

#endif
